@@ -5,7 +5,7 @@ import os
 import re
 import sys
 import uuid
-from typing import Any, Dict, Iterable, List, Union
+from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, cast
 
 from omegaconf import DictConfig, MissingMandatoryValue, OmegaConf
 
@@ -33,6 +33,13 @@ SECRET_MASKS = [
     re.compile(r'.*(password|secret|key)(?!_file)(_|$)').match,
 ]
 
+"""Map of default values"""
+_DEFAULTS: Dict[str, Any] = {
+    'configurations': [],
+    'helpers': {},
+    'testing_configurations': [],
+}
+
 #######################################
 # APPLICATION
 
@@ -44,12 +51,14 @@ class Application:
         name, version, short_description, description, etc.
     """
 
+    __config: Optional[DictConfig]
+
     def __init__(self, **properties) -> None:
         """Initialize the application.
 
         Properties:
         - name: the name of the application (always updated)
-        - verison: version number
+        - version: version number
         - description: description shown in help
         - short_description: shorter description
 
@@ -59,8 +68,15 @@ class Application:
         if not properties.get('name'):
             properties['name'] = self.__get_default_name()
         self.properties = properties
-        self._arg_parser = arg_parser.ArgumentParser(properties)
-        arg_parser.add_default_option_handlers(self._arg_parser)
+        # Add argument parser
+        self.__initialize_parser()
+
+    def __initialize_parser(self):
+        self.parsed = None
+        self.argument_parser = arg_parser.ArgumentParser()
+        parser = self.argument_parser
+        arg_parser.configure_parser(parser, app=self)
+        parser.help_messages.update(_DEFAULTS['helpers'])
 
     @staticmethod
     def __get_default_name() -> str:
@@ -90,17 +106,12 @@ class Application:
         """Get the name of the application"""
         return self.properties['name']
 
-    @property
-    def argument_parser(self):
-        """The argument parser instance"""
-        return self._arg_parser
-
     @property
     def configuration(self) -> DictConfig:
         """Get the configuration of the application, initialize if necessary"""
         if self.__config is None:
             self.setup_configuration(
-                arguments=None, resolve_configuration=False, setup_logging=False
+                arguments=False, resolve_configuration=False, setup_logging=False
             )
             _log.info('alphaconf initialized')
             assert self.__config is not None
@@ -142,7 +153,7 @@ class Application:
             if path and '$' not in path:
                 yield path
 
-    def _load_dotenv(self, load_dotenv: Union[bool, None] = None):
+    def _load_dotenv(self, load_dotenv: Optional[bool] = None):
         """Load dotenv variables (optionally)"""
         if load_dotenv is False:
             return
@@ -152,7 +163,7 @@ class Application:
             _log.debug('Loading dotenv')
             dotenv.load_dotenv()
         except ModuleNotFoundError:
-            if not load_dotenv:
+            if load_dotenv:
                 raise
             _log.debug('dotenv is not installed')
 
@@ -171,17 +182,23 @@ class Application:
         :return: OmegaConf configurations (to be merged)
         """
         _log.debug('Loading default and app configurations')
-        yield from _DEFAULT_CONFIGURATIONS
+        yield from _DEFAULTS['configurations']
+        yield from _DEFAULTS['testing_configurations']
         yield self._app_configuration()
         # Read files
         for path in self._get_possible_configuration_paths():
             if os.path.exists(path):
                 _log.debug('Load configuration from %s', path)
-                yield OmegaConf.load(path)
+                conf = OmegaConf.load(path)
+                if isinstance(conf, DictConfig):
+                    yield conf
+                else:
+                    yield from conf
         # Environment
+        prefixes: Optional[Tuple[str, ...]]
         if env_prefixes is True:
             _log.debug('Detecting accepted env prefixes')
-            default_keys = {k for cfg in _DEFAULT_CONFIGURATIONS for k in cfg.keys()}
+            default_keys = {k for cfg in _DEFAULTS['configurations'] for k in cfg.keys()}
             prefixes = tuple(
                 k.upper() + '_'
                 for k in default_keys
@@ -205,7 +222,7 @@ class Application:
         self,
         arguments: Union[bool, List[str]] = True,
         *,
-        load_dotenv: Union[bool, None] = None,
+        load_dotenv: Optional[bool] = None,
         env_prefixes: Union[bool, Iterable[str]] = True,
         resolve_configuration: bool = True,
         setup_logging: bool = True,
@@ -226,31 +243,22 @@ class Application:
         _log.debug('Start setup application')
 
         # Parse arguments
-        parser_result = None
         if arguments is True:
             arguments = sys.argv[1:]
-        if isinstance(arguments, list):
-            self.argument_parser.reset()
-            self.argument_parser.parse_arguments(arguments)
-            parser_result = self.argument_parser.parse_result
-            _log.debug('Parse arguments result: %s', parser_result)
+        if not isinstance(arguments, list):
+            arguments = []
+        self.parsed = self.argument_parser.parse_args(arguments)
 
         # Load and merge configurations
         self._load_dotenv(load_dotenv=load_dotenv)
         configurations = list(self._get_configurations(env_prefixes=env_prefixes))
-        if parser_result:
-            configurations.extend(self.argument_parser.configurations())
-        self.__config = OmegaConf.merge(*configurations)
+        if self.parsed:
+            configurations.extend(self.parsed.configurations())
+        self.__config = cast(DictConfig, OmegaConf.merge(*configurations))
         _log.debug('Merged %d configurations', len(configurations))
 
         # Handle the result
-        if parser_result == 'show_configuration':
-            print(self.yaml_configuration())
-            raise ExitApplication
-        elif parser_result == 'exit':
-            raise ExitApplication
-        elif parser_result is not None and parser_result != 'ok':
-            raise RuntimeError('Invalid argument parsing result: %s' % parser_result)
+        self._handle_result()
 
         # Try to get the whole configuration to resolve links
         if resolve_configuration:
@@ -259,9 +267,17 @@ class Application:
         # Logging
         if setup_logging:
             _log.debug('Setup logging')
-            self.setup_logging()
+            self._setup_logging()
 
-    def setup_logging(self) -> None:
+    def _handle_result(self):
+        """Handle result that is in self.parsed"""
+        if self.parsed.result:
+            return self.parsed.result.run(self)
+        if self.parsed.rest:
+            raise arg_parser.ArgumentError(f"Too many arguments {self.parsed.rest}")
+        return None
+
+    def _setup_logging(self) -> None:
         """Setup logging
 
         Set the time to GMT, log key 'logging' from configuration or if none, base logging.
@@ -294,7 +310,7 @@ class Application:
         """
         current_config = self.configuration
         try:
-            self.__config = OmegaConf.merge(current_config, conf)
+            self.__config = cast(DictConfig, OmegaConf.merge(current_config, conf))
             yield self
         finally:
             self.__config = current_config
@@ -319,34 +335,61 @@ class Application:
     @staticmethod
     def __mask_secrets(configuration):
         for key in list(configuration):
-            if any(mask(key) for mask in SECRET_MASKS):
+            if isinstance(key, str) and any(mask(key) for mask in SECRET_MASKS):
                 configuration[key] = '*****'
             elif isinstance(configuration[key], (Dict, DictConfig)):
                 configuration[key] = Application.__mask_secrets(configuration[key])
         return configuration
 
-    def run(self, main, should_exit=True, **configuration):
+    def print_help(self, *, usage=None, description=None, arguments=True):
+        """Print the help message
+        Set the arguments to False to disable printing them."""
+        p = self.properties
+        if usage is None:
+            usage = f"usage: {p.get('name') or 'app'}"
+            usage += " [arguments] [key=value ...]"
+        if isinstance(usage, str):
+            print(usage)
+        if description is None:
+            description = p.get('description')
+        if isinstance(description, str):
+            print()
+            print(description)
+        if arguments:
+            print()
+            self.argument_parser.print_help()
+
+    def run(self, main, arguments=True, *, should_exit=True, **configuration):
         """Run this application
 
         :param main: The main function to call
-        :param should_exit: Whether a setup exception should sys.exit (default: True)
+        :param arguments: List of arguments (default: True to read sys.argv)
+        :param should_exit: Whether an exception should sys.exit (default: True)
         :param configuration: Arguments passed to setup_configuration()
         :return: The result of main
         """
         try:
-            self.setup_configuration(**configuration)
+            self.setup_configuration(arguments, **configuration)
         except MissingMandatoryValue as e:
+            _log.error(e)
+            if should_exit:
+                sys.exit(99)
+            raise
+        except arg_parser.ArgumentError as e:
             _log.error(e)
             if should_exit:
                 sys.exit(2)
             raise
-        except ExitApplication:
+        except arg_parser.ExitApplication:
             _log.debug('Normal application exit')
             if should_exit:
                 sys.exit()
             return
-        # Run the application
         log = logging.getLogger()
+        if _DEFAULTS['testing_configurations']:
+            log.info('Application testing (%s: %s)', self.name, main.__qualname__)
+            return None
+        # Run the application
         token = application.set(self)
         try:
             log.info('Application start (%s: %s)', self.name, main.__qualname__)
@@ -358,7 +401,10 @@ class Application:
             return result
         except Exception as e:
             # no need to log exc_info beacause the parent will handle it
-            log.error('Application failed: %s', e)
+            log.error('Application failed (%s) %s', type(e).__name__, e, exc_info=should_exit)
+            if should_exit:
+                log.debug('Exit application')
+                sys.exit(1)
             raise
         finally:
             application.reset(token)
@@ -369,19 +415,12 @@ class Application:
         return f"{type(self).__name__}({self.name}; loaded={ready}; running={running})"
 
 
-class ExitApplication(BaseException):
-    """Signal to exit the application normally"""
-
-    pass
-
-
 #######################################
 # APPLICATION CONTEXT
 
 
 """The application context"""
-application = contextvars.ContextVar('application')
-_DEFAULT_CONFIGURATIONS = []
+application: contextvars.ContextVar[Application] = contextvars.ContextVar('application')
 
 
 def configuration() -> DictConfig:
@@ -396,11 +435,42 @@ def get(config_key: str, type=None) -> Any:
     return app.get_config(config_key, type=type)
 
 
-def setup_configuration(conf: Union[DictConfig, str, Dict]):
-    """Add a default configuration"""
-    if not isinstance(conf, DictConfig):
-        conf = OmegaConf.create(conf)
-    _DEFAULT_CONFIGURATIONS.append(conf)
+def setup_configuration(
+    conf: Union[DictConfig, str, Dict],
+    helpers: Dict[str, str] = {},
+    testing: Optional[bool] = None,
+):
+    """Add a default configuration
+
+    :param conf: The configuration to add
+    :param helpers: Description of parameters used in argument parser helpers
+    :param testing: If set, True adds the configuration to testing configurations,
+                    if False, the testing configurations are cleared
+    """
+    if isinstance(conf, DictConfig):
+        config = conf
+    else:
+        config = cast(DictConfig, OmegaConf.create(conf))
+    if testing is False:
+        _DEFAULTS['testing_configurations'].clear()
+    config_key = 'testing_configurations' if testing else 'configurations'
+    _DEFAULTS[config_key].append(config)
+    # setup helpers
+    for h_key in helpers:
+        key = h_key.split('.', 1)[0]
+        if key not in config:
+            raise ValueError('Invalid helper not in configuration [%s]' % key)
+    _DEFAULTS['helpers'].update(helpers)
+
+
+def check_application():
+    """Check if an application is set up and set it up if it's not"""
+    app = application.get(None)
+    if app is not None:
+        return
+    app = Application()
+    application.set(app)
+    app.setup_configuration(arguments=False, resolve_configuration=False)
 
 
 #######################################
